import logging
import time
from resource import getpagesize

from .process import Process, memory_total_size

columns = [
    "pid",
    "database",
    "client",
    "duration",
    "wait",
    "user",
    "application_name",
    "state",
    "query",
    "iow",
    "read_s",
    "write_s",
    "cpu",
    "memory",
]
logger = logging.getLogger(__name__)


def get_activity(conn, limit):
    """
    Returns PostgreSQL backend list based on pg_stat_activity view.
    For each backend (process) we need to compute: CPU and mem. usage, I/O
    infos.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()
    if conn.server_version >= 90600 and conn.server_version < 100000:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.wait_event_type IS
    DISTINCT FROM 'Lock' THEN 'N' ELSE 'Y' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query,
  pg_stat_activity.application_name as application_name
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """
    elif conn.server_version < 90600:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.waiting = 't' THEN 'Y' ELSE 'N' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query,
  pg_stat_activity.application_name as application_name
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """
    elif conn.server_version >= 100000:
        query = """
SELECT
  pg_stat_activity.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.client_addr AS client,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric, 2)::FLOAT AS duration,
  CASE WHEN pg_stat_activity.wait_event_type IS
    DISTINCT FROM 'Lock' THEN 'N' ELSE 'Y' END AS wait,
  pg_stat_activity.usename AS user,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query,
  pg_stat_activity.application_name as application_name
FROM
  pg_stat_activity
WHERE
  pid <> pg_backend_pid()
  AND backend_type = 'client backend'
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
        """

    query = query + " LIMIT %d" % limit
    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "client": row["client"],
                    "duration": row["duration"],
                    "wait": row["wait"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "state": row["state"],
                    "query": row["query"],
                    "process": Process(row["pid"], mem_total, page_size),
                }
            )
        except Exception as e:
            logger.debug("Failed to process activity row: %s", e)

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row["process"].io_usage()
            if row["duration"] < 0:
                row["duration"] = 0
            final_backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "client": row["client"],
                    "duration": row["duration"],
                    "wait": row["wait"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "state": row["state"],
                    "query": row["query"],
                    "iow": row["process"].iow,
                    "read_s": read_s,
                    "write_s": write_s,
                    "cpu": row["process"].cpu_usage(),
                    "memory": row["process"].mem_usage(),
                }
            )
        except Exception as e:
            logger.debug("Failed to process activity row: %s", e)
    return {"rows": final_backend_list, "columns": columns}


def get_activity_waiting(conn, limit):
    """
    Returns the list of waiting (on lock) queries.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()

    query = """
SELECT
  pg_locks.pid AS pid,
  pg_stat_activity.datname AS database,
  pg_stat_activity.usename AS user,
  pg_locks.mode AS mode,
  pg_locks.locktype AS type,
  COALESCE(pg_locks.relation::regclass::text, ' ') AS relation,
  round(EXTRACT(epoch FROM (NOW()
    - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
  pg_stat_activity.state AS state,
  pg_stat_activity.query AS query,
  pg_stat_activity.application_name as application_name
FROM
  pg_catalog.pg_locks JOIN pg_catalog.pg_stat_activity
    ON (pg_catalog.pg_locks.pid = pg_catalog.pg_stat_activity.pid)
WHERE
  NOT pg_catalog.pg_locks.granted
  AND pg_catalog.pg_stat_activity.pid <> pg_backend_pid()
ORDER BY
  EXTRACT(epoch FROM (NOW() - pg_stat_activity.query_start)) DESC
    """
    query = query + " LIMIT %d" % limit
    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "mode": row["mode"],
                    "type": row["type"],
                    "relation": row["relation"],
                    "duration": row["duration"],
                    "state": row["state"],
                    "query": row["query"],
                    "process": Process(row["pid"], mem_total, page_size),
                }
            )
        except Exception:
            pass

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row["process"].io_usage()
            if row["duration"] < 0:
                row["duration"] = 0
            final_backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "mode": row["mode"],
                    "type": row["type"],
                    "relation": row["relation"],
                    "duration": row["duration"],
                    "state": row["state"],
                    "query": row["query"],
                    "iow": row["process"].iow,
                    "read_s": read_s,
                    "write_s": write_s,
                    "cpu": row["process"].cpu_usage(),
                    "memory": row["process"].mem_usage(),
                }
            )
        except Exception:
            pass
    return {"rows": final_backend_list, "columns": columns}


def get_activity_blocking(conn, limit):
    """
    Returns the list of blocking (lock) queries.
    """
    mem_total = memory_total_size()
    page_size = getpagesize()

    query = """
SELECT
  pid,
  datname AS database,
  usename AS user,
  COALESCE(relation::text, ' ') AS relation,
  mode,
  locktype AS type,
  duration,
  state,
  query,
  application_name
FROM (
  SELECT
    blocking.pid,
    pg_stat_activity.query,
    blocking.mode,
    pg_stat_activity.datname,
    pg_stat_activity.usename,
    blocking.locktype,
    round(EXTRACT(epoch FROM (NOW()
      - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
    blocking.relation::regclass AS relation,
    pg_stat_activity.state AS state,
    pg_stat_activity.application_name as application_name
  FROM
    pg_locks AS blocking
    JOIN (SELECT transactionid FROM pg_locks WHERE NOT granted) AS blocked
      ON (blocking.transactionid = blocked.transactionid)
    JOIN pg_stat_activity
      ON (blocking.pid = pg_stat_activity.pid)
  WHERE
    blocking.granted
  UNION ALL
  SELECT
    blocking.pid,
    pg_stat_activity.query,
    blocking.mode,
    pg_stat_activity.datname,
    pg_stat_activity.usename,
    blocking.locktype,
    round(EXTRACT(epoch FROM (NOW()
      - pg_stat_activity.query_start))::numeric,2)::FLOAT AS duration,
    blocking.relation::regclass AS relation,
    pg_stat_activity.state AS state,
    pg_stat_activity.application_name as application_name
  FROM
    pg_locks AS blocking
    JOIN (SELECT database, relation, mode FROM pg_locks WHERE NOT granted
        AND relation IS NOT NULL) AS blocked
      ON (blocking.database = blocked.database
        AND blocking.relation = blocked.relation)
    JOIN pg_stat_activity
      ON (blocking.pid = pg_stat_activity.pid)
  WHERE
    blocking.granted
) AS sq
ORDER BY duration DESC
    """
    query = query + " LIMIT %d" % limit
    backend_list = []
    for row in conn.query(query):
        try:
            backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "mode": row["mode"],
                    "type": row["type"],
                    "relation": row["relation"],
                    "duration": row["duration"],
                    "state": row["state"],
                    "query": row["query"],
                    "process": Process(row["pid"], mem_total, page_size),
                }
            )
        except Exception:
            pass

    time.sleep(0.1)
    final_backend_list = []
    for row in backend_list:
        try:
            (read_s, write_s) = row["process"].io_usage()
            if row["duration"] < 0:
                row["duration"] = 0
            final_backend_list.append(
                {
                    "pid": row["pid"],
                    "database": row["database"],
                    "user": row["user"],
                    "application_name": row["application_name"],
                    "mode": row["mode"],
                    "type": row["type"],
                    "relation": row["relation"],
                    "duration": row["duration"],
                    "state": row["state"],
                    "query": row["query"],
                    "iow": row["process"].iow,
                    "read_s": read_s,
                    "write_s": write_s,
                    "cpu": row["process"].cpu_usage(),
                    "memory": row["process"].mem_usage(),
                }
            )
        except Exception:
            pass
    return {"rows": final_backend_list, "columns": columns}



def format(**kw):
    return " ".join([format_attribute(*i) for i in sorted(flatten(**kw))])


def format_attribute(key, value):
    if " " in str(value):
        value = '"%s"' % value
    return "%s=%s" % (key, value)


def flatten(**kw):
    for k, v in kw.items():
        if isinstance(v, dict):
            for kk, vv in flatten(**v):
                yield "%s.%s" % (k, kk), vv
        else:
            yield k, v



import os

from setuptools import __version__ as setuptoolsv
from setuptools import find_packages, setup

# Load version number
__version__ = None
setup_path = os.path.dirname(os.path.realpath(__file__))
exec(open(os.path.join(setup_path, "temboardagent", "version.py")).read())

if setuptoolsv < "1.0":
    __version__ = __version__.replace("+", ".")

setup(
    **dict(
        name="temboard-agent",
        version=__version__,
        description="PostgreSQL Remote Control Agent",
        author="Dalibo",
        author_email="contact@dalibo.com",
        license="PostgreSQL",
        url="https://labs.dalibo.com/temboard",
        classifiers=[
            "Environment :: No Input/Output (Daemon)",
            "Intended Audience :: System Administrators",
            "License :: OSI Approved",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3",
            "Topic :: Database :: Database Engines/Servers",
            "Topic :: System :: Monitoring",
        ],
        python_requires="~=3.5",
        include_package_data=True,
        install_requires=["bottle >= 0.12, <0.13", "cryptography"],
        data_files=[
            (
                "share/temboard-agent/",
                [
                    "share/auto_configure.sh",
                    "share/purge.sh",
                    "share/restart-all.sh",
                    "share/preun.sh",
                ],
            ),
            (
                "share/temboard-agent/quickstart/",
                [
                    "share/temboard-agent_CHANGEME.pem",
                    "share/temboard-agent_CHANGEME.key",
                    "share/temboard-agent_ca_certs_CHANGEME.pem",
                ],
            ),
            ("lib/systemd/system", ["packaging/temboard-agent@.service"]),
        ],
        entry_points={
            "console_scripts": ["temboard-agent = temboardagent.__main__:main"],
            "temboardagent.plugins": [
                "activity = temboardagent.plugins.activity:ActivityPlugin",
                "administration = temboardagent.plugins.administration:AdministrationPlugin",  # noqa
                "dashboard = temboardagent.plugins.dashboard:DashboardPlugin",
                "maintenance = temboardagent.plugins.maintenance:MaintenancePlugin",  # noqa
                "monitoring = temboardagent.plugins.monitoring:MonitoringPlugin",
                "pgconf = temboardagent.plugins.pgconf:PgConfPlugin",
                "statements = temboardagent.plugins.statements:StatementsPlugin",
            ],
        },
        packages=find_packages(),
        long_description=open("README.md", encoding="utf-8").read(),
        long_description_content_type="text/markdown",
    )
)



# Various functions to expose components version

import ctypes
import re


def read_distinfo():  # pragma: nocover
    with open("/etc/os-release") as fo:
        distinfos = parse_lsb_release(fo)
    return distinfos


def parse_lsb_release(lines):
    _assignement_re = re.compile(r"""(?P<variable>[A-Z_]+)="(?P<value>[^"]+)"$""")
    infos = dict()
    for line in lines:
        m = _assignement_re.match(line)
        if not m:
            continue
        infos[m.group("variable")] = m.group("value")
    return infos


def load_libpq():
    __import__("psycopg2")

    # Search for libpq.so path in loaded libraries.
    with open("/proc/self/maps") as fo:
        for line in fo:
            values = line.split()
            path = values[-1]
            if "/libpq" in path:
                break
        else:  # pragma: nocover
            raise Exception("libpq.so not loaded")

    return ctypes.cdll.LoadLibrary(path)


def read_libpq_version():
    # Search libpq version bound to this process.

    try:
        # For psycopg2 2.7+
        from psycopg2.extensions import libpq_version

        return libpq_version()
    except ImportError:
        libpq = load_libpq()
        return libpq.PQlibVersion()


def format_pq_version(version):
    pqnums = [version / 10000, version % 100]
    if version <= 100000:
        pqnums[1:1] = [(version % 10000) / 100]
    return ".".join(str(int(n)) for n in pqnums)



from bottle import default_app


class PgConfPlugin:
    PG_MIN_VERSION = (90500, 9.5)

    def __init__(self, app, **kw):
        self.app = app

    def load(self):
        from .routes import bottle

        default_app().mount("/pgconf/", bottle)



import logging
import os
import sqlite3
import time
from datetime import datetime
from textwrap import dedent

from .errors import NotificationError

logger = logging.getLogger(__name__)


class Notification:
    """Notifications are log messages stored in a sqlite table aimed to keep a
    track of each action impacting PostgreSQL server configuration or
    its behavior.

    For now, they are pushed when one of the following actions is triggered:
    - PostgreSQL reloaded.
    - PostgreSQL setting changed.
    - User login.
    - User session expired.
    - PostgreSQL backend termination.

    Maximum number of record is 1000, FIFO behavior is expected.
    """

    def __init__(self, username, message):
        if isinstance(username, bytes):
            username = username.decode("utf-8")
        if isinstance(message, bytes):
            message = message.decode("utf-8")

        self.time = time.time()
        self.username = username
        self.message = message


class NotificationMgmt:
    @classmethod
    def bootstrap(self, config):
        db_path = os.path.join(config.temboard.home, "core.db")
        with sqlite3.connect(db_path) as conn:
            c = conn.cursor()
            c.execute(
                dedent("""
                    CREATE TABLE IF NOT EXISTS action_logs (
                        time REAL,
                        username TEXT,
                        message TEXT
                    )
                """)
            )
            c.execute(
                dedent("""
                    CREATE INDEX IF NOT EXISTS idx_action_logs_time
                    ON action_logs(time)
                """)
            )

    @classmethod
    def push(self, config, notification):
        logger.info("%s user=%s", notification.message, notification.username)
        try:
            db_path = os.path.join(config.temboard.home, "core.db")
            with sqlite3.connect(db_path) as conn:
                c = conn.cursor()
                c.execute(
                    "INSERT INTO action_logs VALUES (?, ?, ?)",
                    (notification.time, notification.username, notification.message),
                )
                # Purge action_logs, we want to keep only the last 100 messages
                c.execute(
                    dedent("""
                        DELETE FROM action_logs
                        WHERE time NOT IN (
                            SELECT time
                            FROM action_logs
                            ORDER BY time DESC
                            LIMIT 1000
                        )
                    """)
                )

        except sqlite3.Error as e:
            logger.exception(str(e))
            raise NotificationError("Can not push new notification")

    @classmethod
    def get_last_n(self, config, n):
        limit = " LIMIT %s" % n if int(n) > -1 else ""

        try:
            db_path = os.path.join(config.temboard.home, "core.db")
            with sqlite3.connect(db_path) as conn:
                c = conn.cursor()
                c.execute(
                    "SELECT time, username, message FROM action_logs "
                    "ORDER BY time DESC " + limit
                )
                for timestamp, username, message in c.fetchall():
                    yield dict(
                        date=datetime.utcfromtimestamp(int(timestamp)).isoformat(),
                        username=username,
                        message=message,
                    )

        except sqlite3.Error as e:
            logger.exception(str(e))
            raise NotificationError("Can not get last notifications")



from base64 import b64decode, b64encode
from hashlib import sha256
from urllib.parse import quote_plus

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends.openssl import backend as openssl_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding

from .errors import TemboardError
from .utils import ensure_bytes

__all__ = ["InvalidSignature"]


_REQUIRED_HEADERS = {
    "host",
    "x-temboard-date",
    "x-temboard-request-id",
    "x-temboard-user",
}

_REQUIRED_POST_HEADERS = _REQUIRED_HEADERS | {"content-type", "content-length"}


def canonicalize_request(method, path, headers, body=b""):
    method = ensure_bytes(method.upper())
    if b"POST" == method:
        required = _REQUIRED_POST_HEADERS
    else:
        required = _REQUIRED_HEADERS

    canonicalized_headers = {
        name.lower(): value
        for name, value in headers.items()
        if name.lower() in required
    }

    missing_headers = required - set(canonicalized_headers)
    if missing_headers:
        missing_headers = ", ".join(sorted(missing_headers))
        raise TemboardError("Missing headers for request signing: %s" % missing_headers)

    lines = [
        b"%s %s" % (method, ensure_bytes(quote_plus(path, safe="/"))),
        b"",
    ] + sorted(
        [
            b"%s: %s" % (ensure_bytes(name), ensure_bytes(value))
            for name, value in canonicalized_headers.items()
        ]
    )

    if b"POST" == method:
        h = sha256()
        h.update(body)
        lines += [b"", h.hexdigest().encode("ascii")]

    lines.append(b"")  # Final EOF

    return b"\n".join(lines)


def load_private_key(data):
    return serialization.load_pem_private_key(
        data, password=None, backend=openssl_backend
    )


def load_public_key(data):
    return serialization.load_pem_public_key(data, backend=openssl_backend)


HASH = hashes.SHA256()
PADDING = padding.PSS(mgf=padding.MGF1(HASH), salt_length=padding.PSS.MAX_LENGTH)


def sign_v1(private_key, payload):
    signature = private_key.sign(payload, PADDING, HASH)
    signature64 = b64encode(signature)
    return signature64.decode("ascii")


def verify_v1(public_key, signature, payload):
    signature_bin = b64decode(signature)
    public_key.verify(signature_bin, payload, PADDING, HASH)



# Mimic asyncio API with synchronous code.
import logging
import os
import signal

logger = logging.getLogger(__name__)


class Service:
    # Mixin for scheduler and worker pool services
    # Manages sync loop and signals.
    name = None

    def __init__(self, app):
        self.app = app

    def __str__(self):
        return self.name

    # interface for services.run
    def create_loop(self):
        return Loop(self)

    # Interface for SignalMultiplexer
    def sighup_handler(self, *a):
        self.app.reload()


class Loop:
    def __init__(self, service):
        self.service = service
        self.signalmngr = SignalManager()
        self.running = False

    def __repr__(self):
        return "<Loop %s>" % self.service

    # Interface for SignalMultiplexer
    def add_signal_handler(self, sig, handler, *a):
        self.signalmngr[sig] = handler

    def remove_signal_handler(self, sig):
        if sig not in self.signalmngr:
            return
        del self.signalmngr[sig]
        if self.running:
            signal.signal(sig, signal.SIG_DFL)

    # Interface for services.run
    def start(self):
        with self.signalmngr:
            ppid = os.getppid()
            self.running = True
            while self.running:
                if self._stop_with_parent(ppid):
                    break

                self.service.accept()

    def stop(self):
        self.running = False

    def close(self):
        self.running = False

    def _stop_with_parent(self, ppid):
        self.running = os.getppid() == ppid
        if not self.running:
            logger.info("Parent process exited. Exiting. ppid=%s", ppid)
        return not self.running


class SignalManager(dict):
    # Context manager for synchronous signal handling.

    def __enter__(self):
        for sig, handler in self.items():
            signal.signal(sig, handler)
        self._registered = self.keys()

    def __exit__(self, *exc):
        for sig in self._registered:
            signal.signal(sig, signal.SIG_DFL)
        del self._registered



# Introspect Postgres instance, temBoard agent and system
#
# Discover is a stable set of properties identifying a running system. temBoard
# computes an ETag from discover data to ease change detection.
#
# Discover data is cached in a discover.json file. temBoard refreshes this file
# in two cases : start of agent, connection lost.
#

import hashlib
import json
import logging
import os
import socket
import sys
import time
from datetime import datetime
from multiprocessing import cpu_count
from platform import machine, python_version

from .core import workers
from .queries import QUERIES
from .toolkit.errors import UserError
from .toolkit.versions import format_pq_version, read_distinfo, read_libpq_version
from .version import __version__

logger = logging.getLogger(__name__)


class Discover:
    def __init__(self, app):
        self.app = app
        self.path = self.app.config.temboard.home + "/discover.json"
        self.data = dict(postgres={}, system={}, temboard={})
        self.json = None  # bytes
        self.etag = None
        self.file_etag = None
        self.mtime = None
        self.inhibit_observer = False

    def connection_lost(self):
        if self.inhibit_observer:
            return
        # Callback for postgres.ConnectionPool connection lost event.
        logger.info("Queueing discover refresh.")
        discover.defer(self.app)

    def ensure_latest(self):
        if self.mtime != os.stat(self.path).st_mtime:
            logger.debug("Discover file changed.")
            return self.read()
        return self.data

    def read(self):
        logger.debug("Reading discover data from %s.", self.path)
        try:
            fo = open(self.path, "rb")
        except OSError as e:
            logger.debug("Failed to read manifest: %s.", e)
            return self.data

        with fo:
            self.json = fo.read()

        self.etag = self.file_etag = compute_etag(self.json)

        try:
            data = json.loads(self.json.decode("utf-8"))
        except json.JSONDecodeError as e:
            raise UserError("Malformed manifest: %s" % e)

        if not isinstance(data, dict):
            raise UserError("Malformed manifest: not a mapping")

        self.data.update(data)
        self.mtime = os.stat(self.path).st_mtime
        return self.data

    def write(self, fo=None):
        if fo is None:
            if self.etag == self.file_etag:
                logger.debug("Discover file up to date.")
                return
            self.mtime = None

        with fo or open(self.path, "w") as fo:
            fo.write(self.json.decode("utf-8"))

        if self.mtime is None:  # if not sys.stdout.
            logger.debug("Wrote discover.json with ETag %s.", self.etag)
            self.mtime = os.stat(self.path).st_mtime
            self.file_etag = self.etag

    def refresh(self, conn=None):
        logger.debug("Inspecting temBoard and system.")
        d = self.data
        old_postgres = self.data.get("postgres", {})
        d.clear()

        d["postgres"] = {}
        d["system"] = {}
        d["temboard"] = {}

        d["temboard"]["bin"] = sys.argv[0]
        d["temboard"]["configfile"] = self.app.config.temboard.configfile
        d["temboard"]["plugins"] = self.app.config.temboard.plugins

        d["system"]["fqdn"] = self.app.config.temboard.hostname
        collect_versions(d)
        collect_cpu(d)
        collect_memory(d)
        collect_system(d)

        try:
            with self.app.postgres.maybe_connect(conn) as pgconn:
                logger.debug("Inspecting Postgres instance.")
                collect_postgres(d, pgconn)
        except Exception as e:
            logger.error("Failed to collect Postgres data: %s", e)
            d["postgres"] = old_postgres

        # Build JSON to compute ETag.
        json_text = json.dumps(self.data, indent="  ", sort_keys=True) + "\n"
        self.json = json_text.encode("utf-8")
        self.etag = compute_etag(self.json)

        if self.etag != self.file_etag:
            logger.info("Instance discover updated.")
        else:
            logger.debug("Instance discover has not changed.")

        return self.data


def collect_cpu(data):
    s = data["system"]
    s["cpu_count"] = cpu_count()

    with open("/proc/cpuinfo") as fo:
        for line in fo:
            if not line.startswith("model name\t"):
                continue
            _, _, model = line.partition("\t: ")
            s["cpu_model"] = model.rstrip()
            break


def collect_memory(data):
    meminfo = {}
    with open("/proc/meminfo") as fo:
        for line in fo:
            if "kB" not in line:
                continue
            field, value, kb = line.split()
            meminfo[field[:-1]] = int(value) * 1024

    s = data["system"]
    s["memory"] = meminfo["MemTotal"]
    s["swap"] = meminfo["SwapTotal"]
    s["hugepage"] = meminfo["Hugepagesize"]


def collect_postgres(data, conn):
    row = conn.queryone(QUERIES["discover"])
    data["postgres"].update(row)
    data["postgres"]["start_time"] = row["start_time"].strftime("%Y-%m-%dT%H:%M:%S%Z")

    for row in conn.query(QUERIES["discover-settings"]):
        t = row["vartype"]
        v = row["setting"]
        if "integer" == t:
            v = int(v)
        elif "bool" == t:
            v = "on" == v

        u = row["unit"]
        if u is None or "B" == u:
            pass
        elif "8kB" == u:
            v = v * 8 * 1024
        elif "MB" == u:
            v = v * 1024 * 1024
        else:
            raise ValueError("Unsupported unit %s" % u)
        data["postgres"][row["name"]] = v
    pid_file_path = find_pid_file(data["postgres"])
    data["postgres"]["pid"] = get_postmaster_pid(pid_file_path)


def get_postmaster_pid(pid_file):
    """Return the pid of the postmaster process"""
    with open(pid_file) as f:
        return int(f.readline().split()[0])


def find_pid_file(postgres_data):
    """Return the pid file path"""
    if postgres_data["external_pid_file"]:
        return postgres_data["external_pid_file"]
    return "%s/postmaster.pid" % postgres_data["data_directory"]


def collect_system(data):
    uname = os.uname()
    s = data["system"]
    s["os"] = uname.sysname
    s["os_version"] = uname.release
    s["arch"] = machine()
    s["hostname"] = socket.gethostname()
    with open("/proc/uptime") as f:
        uptime_seconds = float(f.readline().split()[0])
        s["start_time"] = datetime.utcfromtimestamp(
            int(time.time() - uptime_seconds)
        ).strftime("%Y-%m-%dT%H:%M:%S%Z")


def collect_versions(data):
    versions = inspect_versions()
    data["temboard"].update(dict(agent_version=versions["temboard"]))

    for k in "bottle", "cryptography", "libpq", "psycopg2", "python":
        data["temboard"][k + "_version"] = versions[k]

    data["temboard"]["pythonbin"] = versions["pythonbin"]
    dist = versions["distname"] + " " + versions["distversion"]
    data["system"]["distribution"] = dist


def inspect_versions():
    from bottle import __version__ as bottle_version
    from cryptography import __version__ as cryptography_version
    from psycopg2 import __version__ as psycopg2_version

    distinfos = read_distinfo()

    return dict(
        temboard=__version__,
        temboardbin=sys.argv[0],
        psycopg2=psycopg2_version,
        python=python_version(),
        pythonbin=sys.executable,
        bottle=bottle_version,
        distname=distinfos["NAME"],
        distversion=distinfos.get("VERSION", "n/a"),
        libpq=format_pq_version(read_libpq_version()),
        cryptography=cryptography_version,
    )


def compute_etag(data):
    h = hashlib.new("sha256")
    h.update(data)
    return h.hexdigest()


@workers.register(pool_size=1)
def discover(app):
    """Refresh discover data."""
    app.discover.ensure_latest()
    app.discover.inhibit_observer = True
    app.discover.refresh()
    app.discover.inhibit_observer = False
    app.discover.write()



import json
import os
import sqlite3
from textwrap import dedent

from ...toolkit.utils import JSONEncoder


def bootstrap(path, dbname):
    """Create SQLite3 database model to history collected data for the
    dashboard.

    To be representative of the very recent server activity, dashboard data
    history should not contain old data, that's why the table is dropped and
    recreated when the agent starts.

    We do not need to keep in the history a lot of data (150 records by
    default) and want it to act as a FIFO queue.
    """

    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("DROP TABLE IF EXISTS metrics")
        c.execute(
            dedent("""
                CREATE TABLE metrics (
                    time REAL PRIMARY KEY,
                    data TEXT
                )
            """)
        )


def add_metric(path, dbname, time, data, keep_limit):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute(
            "INSERT INTO metrics VALUES(?, ?)",
            (time, json.dumps(data, cls=JSONEncoder)),
        )
        # Purge
        c.execute(
            dedent("""
                DELETE FROM metrics
                WHERE time NOT IN (
                    SELECT time FROM metrics ORDER BY time DESC
                    LIMIT ?
                )
            """),
            (keep_limit,),
        )


def get_last_metric(path, dbname):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("SELECT data FROM metrics ORDER BY time DESC LIMIT 1")
        return c.fetchone()


def get_all_metrics(path, dbname):
    with sqlite3.connect(os.path.join(path, dbname)) as conn:
        c = conn.cursor()
        c.execute("SELECT data FROM metrics ORDER BY time ASC")
        return c.fetchall()
